import pytest
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.classes.data.builtin.IntegerType import IntegerType
from asn1PERser.classes.types.constraint import ValueRange, MIN, MAX


def SCHEMA_not_contrained_integer(value):
    class MyInt(IntegerType):
        pass
    return MyInt(value)


def SCHEMA_constrained_integer(value, lowerEndpoint_value, upperEndpoint_value, extensionMarker_value=False):
    class MyInt(IntegerType):
        subtypeSpec = ValueRange(lowerEndpoint_value, upperEndpoint_value, extensionMarker=extensionMarker_value)
    return MyInt(value)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_not_contrained_integer(0), '0100'),
    (SCHEMA_not_contrained_integer(1), '0101'),
    (SCHEMA_not_contrained_integer(8), '0108'),
    (SCHEMA_not_contrained_integer(15), '010F'),
    (SCHEMA_not_contrained_integer(16), '0110'),
    (SCHEMA_not_contrained_integer(64), '0140'),
    (SCHEMA_not_contrained_integer(127), '017F'),
    (SCHEMA_not_contrained_integer(128), '020080'),
    (SCHEMA_not_contrained_integer(129), '020081'),
    (SCHEMA_not_contrained_integer(4000), '020FA0'),
    (SCHEMA_not_contrained_integer(32767), '027FFF'),
    (SCHEMA_not_contrained_integer(32768), '03008000'),
    (SCHEMA_not_contrained_integer(8388607), '037FFFFF'),
    (SCHEMA_not_contrained_integer(8388608), '0400800000'),
    (SCHEMA_not_contrained_integer(25181807), '0401803E6F'),
])
def test_unconstrained_greater_or_equal_zero_integers_are_encoded(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_not_contrained_integer(-25181807), '04FE7FC191'),
    (SCHEMA_not_contrained_integer(-8388608), '03800000'),
    (SCHEMA_not_contrained_integer(-8388607), '03800001'),
    (SCHEMA_not_contrained_integer(-32768), '028000'),
    (SCHEMA_not_contrained_integer(-32767), '028001'),
    (SCHEMA_not_contrained_integer(-4000), '02F060'),
    (SCHEMA_not_contrained_integer(-129), '02FF7F'),
    (SCHEMA_not_contrained_integer(-128), '0180'),
    (SCHEMA_not_contrained_integer(-127), '0181'),
    (SCHEMA_not_contrained_integer(-64), '01C0'),
    (SCHEMA_not_contrained_integer(-16), '01F0'),
    (SCHEMA_not_contrained_integer(-15), '01F1'),
    (SCHEMA_not_contrained_integer(-8), '01F8'),
    (SCHEMA_not_contrained_integer(-1), '01FF'),
])
def test_unconstrained_lower_than_zero_integers_are_encoded(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-8388607, lowerEndpoint_value=-8388607, upperEndpoint_value=-8388607), '00'),
    (SCHEMA_constrained_integer(-1864, lowerEndpoint_value=-1864, upperEndpoint_value=-1864), '00'),
    (SCHEMA_constrained_integer(-56, lowerEndpoint_value=-56, upperEndpoint_value=-56), '00'),
    (SCHEMA_constrained_integer(-3, lowerEndpoint_value=-3, upperEndpoint_value=-3), '00'),
    (SCHEMA_constrained_integer(0, lowerEndpoint_value=0, upperEndpoint_value=0), '00'),
    (SCHEMA_constrained_integer(1, lowerEndpoint_value=1, upperEndpoint_value=1), '00'),
    (SCHEMA_constrained_integer(8, lowerEndpoint_value=8, upperEndpoint_value=8), '00'),
    (SCHEMA_constrained_integer(15, lowerEndpoint_value=15, upperEndpoint_value=15), '00'),
    (SCHEMA_constrained_integer(16, lowerEndpoint_value=16, upperEndpoint_value=16), '00'),
    (SCHEMA_constrained_integer(64, lowerEndpoint_value=64, upperEndpoint_value=64), '00'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=127, upperEndpoint_value=127), '00'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=128, upperEndpoint_value=128), '00'),
    (SCHEMA_constrained_integer(129, lowerEndpoint_value=129, upperEndpoint_value=129), '00'),
    (SCHEMA_constrained_integer(4000, lowerEndpoint_value=4000, upperEndpoint_value=4000), '00'),
    (SCHEMA_constrained_integer(32767, lowerEndpoint_value=32767, upperEndpoint_value=32767), '00'),
    (SCHEMA_constrained_integer(32768, lowerEndpoint_value=32768, upperEndpoint_value=32768), '00'),
    (SCHEMA_constrained_integer(8388607, lowerEndpoint_value=8388607, upperEndpoint_value=8388607), '00'),
    (SCHEMA_constrained_integer(8388608, lowerEndpoint_value=8388608, upperEndpoint_value=8388608), '00'),
    (SCHEMA_constrained_integer(25181807, lowerEndpoint_value=25181807, upperEndpoint_value=25181807, ), '00'),
])
def test_constrained_integer_range_has_the_value_1_so_encoding_bit_field_shall_be_empty(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-8388607, lowerEndpoint_value=-8388607, upperEndpoint_value=-8388606), '00'),
    (SCHEMA_constrained_integer(-8388606, lowerEndpoint_value=-8388607, upperEndpoint_value=-8388606), '80'),
    (SCHEMA_constrained_integer(-1864, lowerEndpoint_value=-1864, upperEndpoint_value=-1863), '00'),
    (SCHEMA_constrained_integer(-1863, lowerEndpoint_value=-1864, upperEndpoint_value=-1863), '80'),
    (SCHEMA_constrained_integer(-56, lowerEndpoint_value=-56, upperEndpoint_value=-55), '00'),
    (SCHEMA_constrained_integer(-55, lowerEndpoint_value=-56, upperEndpoint_value=-55), '80'),
    (SCHEMA_constrained_integer(-3, lowerEndpoint_value=-3, upperEndpoint_value=-2), '00'),
    (SCHEMA_constrained_integer(-2, lowerEndpoint_value=-3, upperEndpoint_value=-2), '80'),
    (SCHEMA_constrained_integer(-1, lowerEndpoint_value=-1, upperEndpoint_value=0), '00'),
    (SCHEMA_constrained_integer(0, lowerEndpoint_value=-1, upperEndpoint_value=0), '80'),
    (SCHEMA_constrained_integer(1, lowerEndpoint_value=1, upperEndpoint_value=2), '00'),
    (SCHEMA_constrained_integer(2, lowerEndpoint_value=1, upperEndpoint_value=2), '80'),
    (SCHEMA_constrained_integer(8, lowerEndpoint_value=8, upperEndpoint_value=9), '00'),
    (SCHEMA_constrained_integer(9, lowerEndpoint_value=8, upperEndpoint_value=9), '80'),
    (SCHEMA_constrained_integer(15, lowerEndpoint_value=15, upperEndpoint_value=16), '00'),
    (SCHEMA_constrained_integer(16, lowerEndpoint_value=15, upperEndpoint_value=16), '80'),
    (SCHEMA_constrained_integer(16, lowerEndpoint_value=16, upperEndpoint_value=17), '00'),
    (SCHEMA_constrained_integer(17, lowerEndpoint_value=16, upperEndpoint_value=17), '80'),
    (SCHEMA_constrained_integer(64, lowerEndpoint_value=64, upperEndpoint_value=65), '00'),
    (SCHEMA_constrained_integer(65, lowerEndpoint_value=64, upperEndpoint_value=65), '80'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=127, upperEndpoint_value=128), '00'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=127, upperEndpoint_value=128), '80'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=128, upperEndpoint_value=129), '00'),
    (SCHEMA_constrained_integer(129, lowerEndpoint_value=128, upperEndpoint_value=129), '80'),
    (SCHEMA_constrained_integer(129, lowerEndpoint_value=129, upperEndpoint_value=130), '00'),
    (SCHEMA_constrained_integer(130, lowerEndpoint_value=129, upperEndpoint_value=130), '80'),
    (SCHEMA_constrained_integer(4000, lowerEndpoint_value=4000, upperEndpoint_value=4001), '00'),
    (SCHEMA_constrained_integer(4001, lowerEndpoint_value=4000, upperEndpoint_value=4001), '80'),
    (SCHEMA_constrained_integer(32767, lowerEndpoint_value=32767, upperEndpoint_value=32768), '00'),
    (SCHEMA_constrained_integer(32768, lowerEndpoint_value=32767, upperEndpoint_value=32768), '80'),
    (SCHEMA_constrained_integer(32768, lowerEndpoint_value=32768, upperEndpoint_value=32769), '00'),
    (SCHEMA_constrained_integer(32769, lowerEndpoint_value=32768, upperEndpoint_value=32769), '80'),
    (SCHEMA_constrained_integer(8388607, lowerEndpoint_value=8388607, upperEndpoint_value=8388608), '00'),
    (SCHEMA_constrained_integer(8388608, lowerEndpoint_value=8388607, upperEndpoint_value=8388608), '80'),
    (SCHEMA_constrained_integer(8388608, lowerEndpoint_value=8388608, upperEndpoint_value=8388609), '00'),
    (SCHEMA_constrained_integer(8388609, lowerEndpoint_value=8388608, upperEndpoint_value=8388609), '80'),
    (SCHEMA_constrained_integer(25181807, lowerEndpoint_value=25181807, upperEndpoint_value=25181808, ), '00'),
    (SCHEMA_constrained_integer(25181808, lowerEndpoint_value=25181807, upperEndpoint_value=25181808, ), '80'),
])
def test_constrained_integer_range_has_value_2(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-8388610, lowerEndpoint_value=-8388610, upperEndpoint_value=-8388608), '00'),
    (SCHEMA_constrained_integer(-8388609, lowerEndpoint_value=-8388610, upperEndpoint_value=-8388608), '40'),
    (SCHEMA_constrained_integer(-8388608, lowerEndpoint_value=-8388610, upperEndpoint_value=-8388608), '80'),
    (SCHEMA_constrained_integer(-1, lowerEndpoint_value=-1, upperEndpoint_value=1), '00'),
    (SCHEMA_constrained_integer(0, lowerEndpoint_value=-1, upperEndpoint_value=1), '40'),
    (SCHEMA_constrained_integer(1, lowerEndpoint_value=-1, upperEndpoint_value=1), '80'),
    (SCHEMA_constrained_integer(25181807, lowerEndpoint_value=25181807, upperEndpoint_value=25181809, ), '00'),
    (SCHEMA_constrained_integer(25181808, lowerEndpoint_value=25181807, upperEndpoint_value=25181809, ), '40'),
    (SCHEMA_constrained_integer(25181809, lowerEndpoint_value=25181807, upperEndpoint_value=25181809, ), '80'),
    (SCHEMA_constrained_integer(-56, lowerEndpoint_value=-56, upperEndpoint_value=-53), '00'),
    (SCHEMA_constrained_integer(-55, lowerEndpoint_value=-56, upperEndpoint_value=-53), '40'),
    (SCHEMA_constrained_integer(-54, lowerEndpoint_value=-56, upperEndpoint_value=-53), '80'),
    (SCHEMA_constrained_integer(-53, lowerEndpoint_value=-56, upperEndpoint_value=-53), 'C0'),
    (SCHEMA_constrained_integer(4000, lowerEndpoint_value=4000, upperEndpoint_value=4003), '00'),
    (SCHEMA_constrained_integer(4001, lowerEndpoint_value=4000, upperEndpoint_value=4003), '40'),
    (SCHEMA_constrained_integer(4002, lowerEndpoint_value=4000, upperEndpoint_value=4003), '80'),
    (SCHEMA_constrained_integer(4003, lowerEndpoint_value=4000, upperEndpoint_value=4003), 'C0'),
])
def test_constrained_integer_range_has_value_3_4(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-2, lowerEndpoint_value=-2, upperEndpoint_value=2), '00'),
    (SCHEMA_constrained_integer(-1, lowerEndpoint_value=-2, upperEndpoint_value=2), '20'),
    (SCHEMA_constrained_integer(0, lowerEndpoint_value=-2, upperEndpoint_value=2), '40'),
    (SCHEMA_constrained_integer(1, lowerEndpoint_value=-2, upperEndpoint_value=2), '60'),
    (SCHEMA_constrained_integer(2, lowerEndpoint_value=-2, upperEndpoint_value=2), '80'),
    (SCHEMA_constrained_integer(3, lowerEndpoint_value=-2, upperEndpoint_value=3), 'A0'),
    (SCHEMA_constrained_integer(4, lowerEndpoint_value=-2, upperEndpoint_value=4), 'C0'),
    (SCHEMA_constrained_integer(5, lowerEndpoint_value=-2, upperEndpoint_value=5), 'E0'),
])
def test_constrained_integer_range_has_value_5_6_7_8(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-100000, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '00'),
    (SCHEMA_constrained_integer(-99999, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '10'),
    (SCHEMA_constrained_integer(-99998, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '20'),
    (SCHEMA_constrained_integer(-99997, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '30'),
    (SCHEMA_constrained_integer(-99996, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '40'),
    (SCHEMA_constrained_integer(-99995, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '50'),
    (SCHEMA_constrained_integer(-99994, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '60'),
    (SCHEMA_constrained_integer(-99993, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '70'),
    (SCHEMA_constrained_integer(-99992, lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '80'),
    (SCHEMA_constrained_integer(-99991, lowerEndpoint_value=-100000, upperEndpoint_value=-99991), '90'),
    (SCHEMA_constrained_integer(-99990, lowerEndpoint_value=-100000, upperEndpoint_value=-99990), 'A0'),
    (SCHEMA_constrained_integer(-99989, lowerEndpoint_value=-100000, upperEndpoint_value=-99989), 'B0'),
    (SCHEMA_constrained_integer(-99988, lowerEndpoint_value=-100000, upperEndpoint_value=-99988), 'C0'),
    (SCHEMA_constrained_integer(-99987, lowerEndpoint_value=-100000, upperEndpoint_value=-99987), 'D0'),
    (SCHEMA_constrained_integer(-99986, lowerEndpoint_value=-100000, upperEndpoint_value=-99986), 'E0'),
    (SCHEMA_constrained_integer(-99985, lowerEndpoint_value=-100000, upperEndpoint_value=-99985), 'F0'),
])
def test_constrained_integer_range_has_value_9_to_16(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-4001, lowerEndpoint_value=-4016, upperEndpoint_value=-4000), '78'),
    (SCHEMA_constrained_integer(4001, lowerEndpoint_value=4000, upperEndpoint_value=4016), '08'),
    (SCHEMA_constrained_integer(4003, lowerEndpoint_value=4000, upperEndpoint_value=4016), '18'),
    (SCHEMA_constrained_integer(4005, lowerEndpoint_value=4000, upperEndpoint_value=4016), '28'),
    (SCHEMA_constrained_integer(4010, lowerEndpoint_value=4000, upperEndpoint_value=4016), '50'),
    (SCHEMA_constrained_integer(4014, lowerEndpoint_value=4000, upperEndpoint_value=4016), '70'),
    (SCHEMA_constrained_integer(4016, lowerEndpoint_value=4000, upperEndpoint_value=4016), '80'),
    (SCHEMA_constrained_integer(4017, lowerEndpoint_value=4000, upperEndpoint_value=4017), '88'),
    (SCHEMA_constrained_integer(4020, lowerEndpoint_value=4000, upperEndpoint_value=4020), 'A0'),
    (SCHEMA_constrained_integer(4024, lowerEndpoint_value=4000, upperEndpoint_value=4024), 'C0'),
    (SCHEMA_constrained_integer(4028, lowerEndpoint_value=4000, upperEndpoint_value=4028), 'E0'),
    (SCHEMA_constrained_integer(4031, lowerEndpoint_value=4000, upperEndpoint_value=4031), 'F8'),
])
def test_constrained_integer_range_has_value_17_to_32(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-1000001, lowerEndpoint_value=-1000032, upperEndpoint_value=-1000000), '7C'),
    (SCHEMA_constrained_integer(1000001, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '04'),
    (SCHEMA_constrained_integer(1000005, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '14'),
    (SCHEMA_constrained_integer(1000009, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '24'),
    (SCHEMA_constrained_integer(1000016, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '40'),
    (SCHEMA_constrained_integer(1000022, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '58'),
    (SCHEMA_constrained_integer(1000023, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '5C'),
    (SCHEMA_constrained_integer(1000024, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '60'),
    (SCHEMA_constrained_integer(1000028, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '70'),
    (SCHEMA_constrained_integer(1000032, lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '80'),
    (SCHEMA_constrained_integer(1000037, lowerEndpoint_value=1000000, upperEndpoint_value=1000040), '94'),
    (SCHEMA_constrained_integer(1000045, lowerEndpoint_value=1000000, upperEndpoint_value=1000045), 'B4'),
    (SCHEMA_constrained_integer(1000055, lowerEndpoint_value=1000000, upperEndpoint_value=1000056), 'DC'),
    (SCHEMA_constrained_integer(1000063, lowerEndpoint_value=1000000, upperEndpoint_value=1000063), 'FC'),
])
def test_constrained_integer_range_has_value_33_to_64(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-7, lowerEndpoint_value=-64, upperEndpoint_value=0), '72'),
    (SCHEMA_constrained_integer(7, lowerEndpoint_value=0, upperEndpoint_value=64), '0E'),
    (SCHEMA_constrained_integer(25, lowerEndpoint_value=0, upperEndpoint_value=64), '32'),
    (SCHEMA_constrained_integer(44, lowerEndpoint_value=0, upperEndpoint_value=64), '58'),
    (SCHEMA_constrained_integer(61, lowerEndpoint_value=0, upperEndpoint_value=64), '7A'),
    (SCHEMA_constrained_integer(69, lowerEndpoint_value=0, upperEndpoint_value=70), '8A'),
    (SCHEMA_constrained_integer(77, lowerEndpoint_value=0, upperEndpoint_value=77), '9A'),
    (SCHEMA_constrained_integer(85, lowerEndpoint_value=0, upperEndpoint_value=85), 'AA'),
    (SCHEMA_constrained_integer(98, lowerEndpoint_value=0, upperEndpoint_value=98), 'C4'),
    (SCHEMA_constrained_integer(111, lowerEndpoint_value=0, upperEndpoint_value=111), 'DE'),
    (SCHEMA_constrained_integer(119, lowerEndpoint_value=0, upperEndpoint_value=120), 'EE'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=0, upperEndpoint_value=127), 'FE'),
])
def test_constrained_integer_range_has_value_65_to_128(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-7, lowerEndpoint_value=-128, upperEndpoint_value=0), '79'),
    (SCHEMA_constrained_integer(7, lowerEndpoint_value=0, upperEndpoint_value=128), '07'),
    (SCHEMA_constrained_integer(25, lowerEndpoint_value=0, upperEndpoint_value=128), '19'),
    (SCHEMA_constrained_integer(44, lowerEndpoint_value=0, upperEndpoint_value=128), '2C'),
    (SCHEMA_constrained_integer(61, lowerEndpoint_value=0, upperEndpoint_value=128), '3D'),
    (SCHEMA_constrained_integer(69, lowerEndpoint_value=0, upperEndpoint_value=128), '45'),
    (SCHEMA_constrained_integer(77, lowerEndpoint_value=0, upperEndpoint_value=128), '4D'),
    (SCHEMA_constrained_integer(85, lowerEndpoint_value=0, upperEndpoint_value=128), '55'),
    (SCHEMA_constrained_integer(98, lowerEndpoint_value=0, upperEndpoint_value=128), '62'),
    (SCHEMA_constrained_integer(111, lowerEndpoint_value=0, upperEndpoint_value=128), '6F'),
    (SCHEMA_constrained_integer(119, lowerEndpoint_value=0, upperEndpoint_value=128), '77'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=0, upperEndpoint_value=128), '7F'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=0, upperEndpoint_value=128), '80'),
    (SCHEMA_constrained_integer(140, lowerEndpoint_value=0, upperEndpoint_value=140), '8C'),
    (SCHEMA_constrained_integer(161, lowerEndpoint_value=0, upperEndpoint_value=161), 'A1'),
    (SCHEMA_constrained_integer(182, lowerEndpoint_value=0, upperEndpoint_value=182), 'B6'),
    (SCHEMA_constrained_integer(230, lowerEndpoint_value=0, upperEndpoint_value=230), 'E6'),
    (SCHEMA_constrained_integer(253, lowerEndpoint_value=0, upperEndpoint_value=253), 'FD'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=254), 'FE'),
])
def test_constrained_integer_range_has_value_129_to_255(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-7, lowerEndpoint_value=-255, upperEndpoint_value=0), 'F8'),
    (SCHEMA_constrained_integer(7, lowerEndpoint_value=0, upperEndpoint_value=255), '07'),
    (SCHEMA_constrained_integer(25, lowerEndpoint_value=0, upperEndpoint_value=255), '19'),
    (SCHEMA_constrained_integer(44, lowerEndpoint_value=0, upperEndpoint_value=255), '2C'),
    (SCHEMA_constrained_integer(61, lowerEndpoint_value=0, upperEndpoint_value=255), '3D'),
    (SCHEMA_constrained_integer(69, lowerEndpoint_value=0, upperEndpoint_value=255), '45'),
    (SCHEMA_constrained_integer(77, lowerEndpoint_value=0, upperEndpoint_value=255), '4D'),
    (SCHEMA_constrained_integer(85, lowerEndpoint_value=0, upperEndpoint_value=255), '55'),
    (SCHEMA_constrained_integer(98, lowerEndpoint_value=0, upperEndpoint_value=255), '62'),
    (SCHEMA_constrained_integer(111, lowerEndpoint_value=0, upperEndpoint_value=255), '6F'),
    (SCHEMA_constrained_integer(119, lowerEndpoint_value=0, upperEndpoint_value=255), '77'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=0, upperEndpoint_value=255), '7F'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=0, upperEndpoint_value=255), '80'),
    (SCHEMA_constrained_integer(140, lowerEndpoint_value=0, upperEndpoint_value=255), '8C'),
    (SCHEMA_constrained_integer(161, lowerEndpoint_value=0, upperEndpoint_value=255), 'A1'),
    (SCHEMA_constrained_integer(182, lowerEndpoint_value=0, upperEndpoint_value=255), 'B6'),
    (SCHEMA_constrained_integer(230, lowerEndpoint_value=0, upperEndpoint_value=255), 'E6'),
    (SCHEMA_constrained_integer(253, lowerEndpoint_value=0, upperEndpoint_value=255), 'FD'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=255), 'FE'),
    (SCHEMA_constrained_integer(255, lowerEndpoint_value=0, upperEndpoint_value=255), 'FF'),
])
def test_constrained_integer_range_has_value_256(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-7, lowerEndpoint_value=-256, upperEndpoint_value=0), '00F9'),
    (SCHEMA_constrained_integer(7, lowerEndpoint_value=0, upperEndpoint_value=256), '0007'),
    (SCHEMA_constrained_integer(25, lowerEndpoint_value=0, upperEndpoint_value=256), '0019'),
    (SCHEMA_constrained_integer(44, lowerEndpoint_value=0, upperEndpoint_value=256), '002C'),
    (SCHEMA_constrained_integer(61, lowerEndpoint_value=0, upperEndpoint_value=256), '003D'),
    (SCHEMA_constrained_integer(69, lowerEndpoint_value=0, upperEndpoint_value=256), '0045'),
    (SCHEMA_constrained_integer(77, lowerEndpoint_value=0, upperEndpoint_value=256), '004D'),
    (SCHEMA_constrained_integer(85, lowerEndpoint_value=0, upperEndpoint_value=256), '0055'),
    (SCHEMA_constrained_integer(98, lowerEndpoint_value=0, upperEndpoint_value=256), '0062'),
    (SCHEMA_constrained_integer(111, lowerEndpoint_value=0, upperEndpoint_value=256), '006F'),
    (SCHEMA_constrained_integer(119, lowerEndpoint_value=0, upperEndpoint_value=256), '0077'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=0, upperEndpoint_value=256), '007F'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=0, upperEndpoint_value=256), '0080'),
    (SCHEMA_constrained_integer(140, lowerEndpoint_value=0, upperEndpoint_value=256), '008C'),
    (SCHEMA_constrained_integer(161, lowerEndpoint_value=0, upperEndpoint_value=256), '00A1'),
    (SCHEMA_constrained_integer(182, lowerEndpoint_value=0, upperEndpoint_value=256), '00B6'),
    (SCHEMA_constrained_integer(230, lowerEndpoint_value=0, upperEndpoint_value=256), '00E6'),
    (SCHEMA_constrained_integer(253, lowerEndpoint_value=0, upperEndpoint_value=256), '00FD'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=256), '00FE'),
    (SCHEMA_constrained_integer(255, lowerEndpoint_value=0, upperEndpoint_value=256), '00FF'),
    (SCHEMA_constrained_integer(256, lowerEndpoint_value=0, upperEndpoint_value=256), '0100'),
    (SCHEMA_constrained_integer(511, lowerEndpoint_value=0, upperEndpoint_value=511), '01FF'),
    (SCHEMA_constrained_integer(512, lowerEndpoint_value=0, upperEndpoint_value=512), '0200'),
    (SCHEMA_constrained_integer(1023, lowerEndpoint_value=0, upperEndpoint_value=1023), '03FF'),
    (SCHEMA_constrained_integer(1024, lowerEndpoint_value=0, upperEndpoint_value=1024), '0400'),
    (SCHEMA_constrained_integer(1024, lowerEndpoint_value=0, upperEndpoint_value=1024), '0400'),
    (SCHEMA_constrained_integer(8191, lowerEndpoint_value=0, upperEndpoint_value=8191), '1FFF'),
    (SCHEMA_constrained_integer(8192, lowerEndpoint_value=0, upperEndpoint_value=8192), '2000'),
    (SCHEMA_constrained_integer(32766, lowerEndpoint_value=0, upperEndpoint_value=32768), '7FFE'),
    (SCHEMA_constrained_integer(32767, lowerEndpoint_value=0, upperEndpoint_value=32768), '7FFF'),
    (SCHEMA_constrained_integer(32768, lowerEndpoint_value=0, upperEndpoint_value=32768), '8000'),
    (SCHEMA_constrained_integer(44321, lowerEndpoint_value=0, upperEndpoint_value=44321), 'AD21'),
    (SCHEMA_constrained_integer(55555, lowerEndpoint_value=0, upperEndpoint_value=60000), 'D903'),
    (SCHEMA_constrained_integer(60000, lowerEndpoint_value=0, upperEndpoint_value=65535), 'EA60'),
    (SCHEMA_constrained_integer(65534, lowerEndpoint_value=0, upperEndpoint_value=65535), 'FFFE'),
    (SCHEMA_constrained_integer(65535, lowerEndpoint_value=0, upperEndpoint_value=65535), 'FFFF'),
])
def test_constrained_integer_range_has_value_257_to_65536(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(-100000, lowerEndpoint_value=-1234567, upperEndpoint_value=0), '80114FE7'),
    (SCHEMA_constrained_integer(7, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0007'),
    (SCHEMA_constrained_integer(25, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0019'),
    (SCHEMA_constrained_integer(44, lowerEndpoint_value=0, upperEndpoint_value=1234567), '002C'),
    (SCHEMA_constrained_integer(61, lowerEndpoint_value=0, upperEndpoint_value=1234567), '003D'),
    (SCHEMA_constrained_integer(69, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0045'),
    (SCHEMA_constrained_integer(77, lowerEndpoint_value=0, upperEndpoint_value=1234567), '004D'),
    (SCHEMA_constrained_integer(85, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0055'),
    (SCHEMA_constrained_integer(98, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0062'),
    (SCHEMA_constrained_integer(111, lowerEndpoint_value=0, upperEndpoint_value=1234567), '006F'),
    (SCHEMA_constrained_integer(119, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0077'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=0, upperEndpoint_value=1234567), '007F'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=0, upperEndpoint_value=1234567), '0080'),
    (SCHEMA_constrained_integer(140, lowerEndpoint_value=0, upperEndpoint_value=1234567), '008C'),
    (SCHEMA_constrained_integer(161, lowerEndpoint_value=0, upperEndpoint_value=1234567), '00A1'),
    (SCHEMA_constrained_integer(182, lowerEndpoint_value=0, upperEndpoint_value=1234567), '00B6'),
    (SCHEMA_constrained_integer(230, lowerEndpoint_value=0, upperEndpoint_value=1234567), '00E6'),
    (SCHEMA_constrained_integer(253, lowerEndpoint_value=0, upperEndpoint_value=1234567), '00FD'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=1234567), '00FE'),
    (SCHEMA_constrained_integer(255, lowerEndpoint_value=0, upperEndpoint_value=1234567), '00FF'),
    (SCHEMA_constrained_integer(256, lowerEndpoint_value=0, upperEndpoint_value=1234567), '400100'),
    (SCHEMA_constrained_integer(511, lowerEndpoint_value=0, upperEndpoint_value=1234567), '4001FF'),
    (SCHEMA_constrained_integer(512, lowerEndpoint_value=0, upperEndpoint_value=1234567), '400200'),
    (SCHEMA_constrained_integer(1023, lowerEndpoint_value=0, upperEndpoint_value=1234567), '4003FF'),
    (SCHEMA_constrained_integer(1024, lowerEndpoint_value=0, upperEndpoint_value=1234567), '400400'),
    (SCHEMA_constrained_integer(1024, lowerEndpoint_value=0, upperEndpoint_value=1234567), '400400'),
    (SCHEMA_constrained_integer(8191, lowerEndpoint_value=0, upperEndpoint_value=1234567), '401FFF'),
    (SCHEMA_constrained_integer(8192, lowerEndpoint_value=0, upperEndpoint_value=1234567), '402000'),
    (SCHEMA_constrained_integer(32766, lowerEndpoint_value=0, upperEndpoint_value=1234567), '407FFE'),
    (SCHEMA_constrained_integer(32767, lowerEndpoint_value=0, upperEndpoint_value=1234567), '407FFF'),
    (SCHEMA_constrained_integer(32768, lowerEndpoint_value=0, upperEndpoint_value=1234567), '408000'),
    (SCHEMA_constrained_integer(44321, lowerEndpoint_value=0, upperEndpoint_value=1234567), '40AD21'),
    (SCHEMA_constrained_integer(55555, lowerEndpoint_value=0, upperEndpoint_value=1234567), '40D903'),
    (SCHEMA_constrained_integer(60000, lowerEndpoint_value=0, upperEndpoint_value=1234567), '40EA60'),
    (SCHEMA_constrained_integer(65534, lowerEndpoint_value=0, upperEndpoint_value=1234567), '40FFFE'),
    (SCHEMA_constrained_integer(65535, lowerEndpoint_value=0, upperEndpoint_value=1234567), '40FFFF'),
    (SCHEMA_constrained_integer(65536, lowerEndpoint_value=0, upperEndpoint_value=1234567), '80010000'),
    (SCHEMA_constrained_integer(100000, lowerEndpoint_value=0, upperEndpoint_value=1234567), '800186A0'),
    (SCHEMA_constrained_integer(256, lowerEndpoint_value=256, upperEndpoint_value=1234567), '0000'),
    (SCHEMA_constrained_integer(100, lowerEndpoint_value=0, upperEndpoint_value=4294967295), '0064'),
    (SCHEMA_constrained_integer(255, lowerEndpoint_value=0, upperEndpoint_value=4294967295), '00FF'),
    (SCHEMA_constrained_integer(256, lowerEndpoint_value=0, upperEndpoint_value=4294967295), '400100'),
    (SCHEMA_constrained_integer(100000, lowerEndpoint_value=0, upperEndpoint_value=4294967295), '800186A0'),
    (SCHEMA_constrained_integer(100000000, lowerEndpoint_value=0, upperEndpoint_value=4294967295), 'C005F5E100'),
    (SCHEMA_constrained_integer(4294967294, lowerEndpoint_value=0, upperEndpoint_value=4294967295), 'C0FFFFFFFE'),
    (SCHEMA_constrained_integer(4294967295, lowerEndpoint_value=0, upperEndpoint_value=4294967295), 'C0FFFFFFFF'),
])
def test_constrained_integer_range_has_value_greater_than_65536(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(7, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0107'),
    (SCHEMA_constrained_integer(25, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0119'),
    (SCHEMA_constrained_integer(44, lowerEndpoint_value=0, upperEndpoint_value=MAX), '012C'),
    (SCHEMA_constrained_integer(61, lowerEndpoint_value=0, upperEndpoint_value=MAX), '013D'),
    (SCHEMA_constrained_integer(69, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0145'),
    (SCHEMA_constrained_integer(77, lowerEndpoint_value=0, upperEndpoint_value=MAX), '014D'),
    (SCHEMA_constrained_integer(85, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0155'),
    (SCHEMA_constrained_integer(98, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0162'),
    (SCHEMA_constrained_integer(111, lowerEndpoint_value=0, upperEndpoint_value=MAX), '016F'),
    (SCHEMA_constrained_integer(119, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0177'),
    (SCHEMA_constrained_integer(127, lowerEndpoint_value=0, upperEndpoint_value=MAX), '017F'),
    (SCHEMA_constrained_integer(128, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0180'),
    (SCHEMA_constrained_integer(140, lowerEndpoint_value=0, upperEndpoint_value=MAX), '018C'),
    (SCHEMA_constrained_integer(161, lowerEndpoint_value=0, upperEndpoint_value=MAX), '01A1'),
    (SCHEMA_constrained_integer(182, lowerEndpoint_value=0, upperEndpoint_value=MAX), '01B6'),
    (SCHEMA_constrained_integer(230, lowerEndpoint_value=0, upperEndpoint_value=MAX), '01E6'),
    (SCHEMA_constrained_integer(253, lowerEndpoint_value=0, upperEndpoint_value=MAX), '01FD'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=MAX), '01FE'),
    (SCHEMA_constrained_integer(255, lowerEndpoint_value=0, upperEndpoint_value=MAX), '01FF'),
    (SCHEMA_constrained_integer(256, lowerEndpoint_value=0, upperEndpoint_value=MAX), '020100'),
    (SCHEMA_constrained_integer(511, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0201FF'),
    (SCHEMA_constrained_integer(512, lowerEndpoint_value=0, upperEndpoint_value=MAX), '020200'),
    (SCHEMA_constrained_integer(1023, lowerEndpoint_value=0, upperEndpoint_value=MAX), '0203FF'),
    (SCHEMA_constrained_integer(1024, lowerEndpoint_value=0, upperEndpoint_value=MAX), '020400'),
    (SCHEMA_constrained_integer(1024, lowerEndpoint_value=0, upperEndpoint_value=MAX), '020400'),
    (SCHEMA_constrained_integer(8191, lowerEndpoint_value=0, upperEndpoint_value=MAX), '021FFF'),
    (SCHEMA_constrained_integer(8192, lowerEndpoint_value=0, upperEndpoint_value=MAX), '022000'),
    (SCHEMA_constrained_integer(32766, lowerEndpoint_value=0, upperEndpoint_value=MAX), '027FFE'),
    (SCHEMA_constrained_integer(32767, lowerEndpoint_value=0, upperEndpoint_value=MAX), '027FFF'),
    (SCHEMA_constrained_integer(32768, lowerEndpoint_value=0, upperEndpoint_value=MAX), '028000'),
    (SCHEMA_constrained_integer(44321, lowerEndpoint_value=0, upperEndpoint_value=MAX), '02AD21'),
    (SCHEMA_constrained_integer(55555, lowerEndpoint_value=0, upperEndpoint_value=MAX), '02D903'),
    (SCHEMA_constrained_integer(60000, lowerEndpoint_value=0, upperEndpoint_value=MAX), '02EA60'),
    (SCHEMA_constrained_integer(65534, lowerEndpoint_value=0, upperEndpoint_value=MAX), '02FFFE'),
    (SCHEMA_constrained_integer(65535, lowerEndpoint_value=0, upperEndpoint_value=MAX), '02FFFF'),
    (SCHEMA_constrained_integer(65536, lowerEndpoint_value=0, upperEndpoint_value=MAX), '03010000'),
    (SCHEMA_constrained_integer(100000, lowerEndpoint_value=0, upperEndpoint_value=MAX), '030186A0'),
    (SCHEMA_constrained_integer(256, lowerEndpoint_value=256, upperEndpoint_value=MAX), '0100')
])
def test_semi_constrained_integer_is_encoded(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(3, lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '60'),
    (SCHEMA_constrained_integer(3, lowerEndpoint_value=0, upperEndpoint_value=120, extensionMarker_value=True), '03'),
    (SCHEMA_constrained_integer(3, lowerEndpoint_value=0, upperEndpoint_value=255, extensionMarker_value=True), '0003'),
    (SCHEMA_constrained_integer(3, lowerEndpoint_value=0, upperEndpoint_value=10000, extensionMarker_value=True), '000003'),
    (SCHEMA_constrained_integer(3, lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '0003'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=255, extensionMarker_value=True), '00FE'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=10000, extensionMarker_value=True), '0000FE'),
    (SCHEMA_constrained_integer(254, lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '00FE'),
    (SCHEMA_constrained_integer(1023, lowerEndpoint_value=0, upperEndpoint_value=10000, extensionMarker_value=True), '0003FF'),
    (SCHEMA_constrained_integer(1023, lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '2003FF'),
    (SCHEMA_constrained_integer(32766, lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '207FFE'),
    (SCHEMA_constrained_integer(65535, lowerEndpoint_value=0, upperEndpoint_value=1000000, extensionMarker_value=True), '20FFFF'),
    (SCHEMA_constrained_integer(100000, lowerEndpoint_value=0, upperEndpoint_value=1000000, extensionMarker_value=True), '400186A0'),
])
def test_extension_marker_is_present_and_value_is_within_extension_root_integer_is_encoded(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("integer, encoded", [
    (SCHEMA_constrained_integer(4, lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '800104'),
    (SCHEMA_constrained_integer(100000, lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '80030186A0'),
    (SCHEMA_constrained_integer(-4, lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '8001FC'),
    (SCHEMA_constrained_integer(-100000, lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '8003FE7960'),
])
def test_extension_marker_is_present_and_values_in_not_within_extension_root_integer_is_encoded(integer, encoded):
    assert per_encoder(integer) == bytearray.fromhex(encoded)



from pyasn1.type.univ import OctetString
from asn1PERser.codec.per.encoder import encode_octetstring
from asn1PERser.codec.per.decoder import decode_octetstring
from asn1PERser.classes.types.constraint import NoConstraint


class OctetStringType(OctetString):
    subtypeSpec = NoConstraint()

    def fill_field_list(self, field_list):
        octetstring_field_list = encode_octetstring(self)
        if octetstring_field_list:
            field_list.extend(octetstring_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_octetstring(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        return ''.join(format(byte, '02x') for byte in self.asNumbers())



import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_octetstring import SCHEMA_constrained_ext_octetstring, \
    SCHEMA_constrained_octetstring, SCHEMA_no_constrain_octetstring, DATA_octetstring


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_no_constrain_octetstring(), '0180', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='1')),
    (SCHEMA_no_constrain_octetstring(), '0100', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='0')),
    (SCHEMA_no_constrain_octetstring(), '0140', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='01')),
    (SCHEMA_no_constrain_octetstring(), '0180', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='10')),
    (SCHEMA_no_constrain_octetstring(), '01F0', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='1111')),
    (SCHEMA_no_constrain_octetstring(), '01F0', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='111100')),
    (SCHEMA_no_constrain_octetstring(), '02F0F0', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='1111000011110')),
    (SCHEMA_no_constrain_octetstring(), '020C33', DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='0000110000110011')),
    (SCHEMA_no_constrain_octetstring(), '075CB35B6E395CA8', DATA_octetstring(SCHEMA_no_constrain_octetstring(),
                                                                             bin_val='01011100101100110101101101101110001110010101110010101')),
])
def test_no_constrain_octetstring_as_binary_value_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_no_constrain_octetstring(), '0100', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='0')),
    (SCHEMA_no_constrain_octetstring(), '0110', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='1')),
    (SCHEMA_no_constrain_octetstring(), '0101', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='01')),
    (SCHEMA_no_constrain_octetstring(), '0110', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='10')),
    (SCHEMA_no_constrain_octetstring(), '0200A0', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='00A')),
    (SCHEMA_no_constrain_octetstring(), '02AA00', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='AA0')),
    (SCHEMA_no_constrain_octetstring(), '04ABCDEFA0', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='ABCDEFA')),
    (SCHEMA_no_constrain_octetstring(), '080123ABCDEFA01230', DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='0123ABCDEFA0123')),
])
def test_no_constrain_octetstring_as_hex_value_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    # (SCHEMA_constrained_octetstring(lb=0, ub=0), '00', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=0), bin_val='')),    # does not work
    (SCHEMA_constrained_octetstring(lb=0, ub=0), '00', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=0), hex_val='')),
])
def test_constrained_octetstring_of_zero_length_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_octetstring(lb=1, ub=1), '80', DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), bin_val='1')),
    (SCHEMA_constrained_octetstring(lb=1, ub=1), '70', DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), bin_val='01110')),
    (SCHEMA_constrained_octetstring(lb=1, ub=1), '0F', DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), bin_val='00001111')),
    (SCHEMA_constrained_octetstring(lb=1, ub=1), 'C0', DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), hex_val='C')),
    (SCHEMA_constrained_octetstring(lb=1, ub=1), 'CF', DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), hex_val='CF')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), 'FF00', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), bin_val='111111110')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), '0F08', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), bin_val='0000111100001')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), 'CCCC', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), bin_val='1100110011001100')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), 'ABC0', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='ABC')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), '0000', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='000')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), '0000', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='0000')),
    (SCHEMA_constrained_octetstring(lb=2, ub=2), 'FFFF', DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='FFFF')),
])
def test_constrained_same_len_octetstring_with_len_le_2_octetes_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_octetstring(lb=3, ub=3), '010101', DATA_octetstring(SCHEMA_constrained_octetstring(lb=3, ub=3), bin_val='000000010000000100000001')),
    (SCHEMA_constrained_octetstring(lb=3, ub=3), 'ABCDEF', DATA_octetstring(SCHEMA_constrained_octetstring(lb=3, ub=3), hex_val='ABCDEF')),
    (SCHEMA_constrained_octetstring(lb=4, ub=4), '7FFFFF00', DATA_octetstring(SCHEMA_constrained_octetstring(lb=4, ub=4), bin_val='01111111111111111111111100000000')),
    (SCHEMA_constrained_octetstring(lb=4, ub=4), 'ABCDEF01', DATA_octetstring(SCHEMA_constrained_octetstring(lb=4, ub=4), hex_val='ABCDEF01')),
    (SCHEMA_constrained_octetstring(lb=17, ub=17), '11223344556677889900AABBCCDDEEFF11', DATA_octetstring(SCHEMA_constrained_octetstring(lb=17, ub=17),
                                                                                                          hex_val='11223344556677889900AABBCCDDEEFF11')),
])
def test_constrained_same_len_octetstring_with_len_gt_2_and_lt_64K_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_octetstring(lb=0, ub=15), '00', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val='')),
    (SCHEMA_constrained_octetstring(lb=0, ub=15), '10A0', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val='A')),
    (SCHEMA_constrained_octetstring(lb=0, ub=15), '10A0', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val='A0')),
    (SCHEMA_constrained_octetstring(lb=0, ub=15), 'F0112233445566778899001122334455', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15),
                                                                                                       hex_val='112233445566778899001122334455')),
    (SCHEMA_constrained_octetstring(lb=0, ub=22), '08A0', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22), hex_val='A')),
    (SCHEMA_constrained_octetstring(lb=0, ub=22), '08A0', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22), hex_val='A0')),
    (SCHEMA_constrained_octetstring(lb=0, ub=22), '78112233445566778899001122334455', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22),
                                                                                                       hex_val='112233445566778899001122334455')),
    (SCHEMA_constrained_octetstring(lb=0, ub=22), 'B011223344556677889900112233445566778899001122', DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22),
                                                                                                                     hex_val='11223344556677889900112233445566778899001122')),
    (SCHEMA_constrained_octetstring(lb=10, ub=15), '00112233445566778899AA', DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=15),
                                                                                              hex_val='112233445566778899AA')),
    (SCHEMA_constrained_octetstring(lb=10, ub=15), 'A0112233445566778899AABBCCDDEEFF', DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=15),
                                                                                                        hex_val='112233445566778899AABBCCDDEEFF')),
    (SCHEMA_constrained_octetstring(lb=10, ub=22), '00112233445566778899AA', DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=22),
                                                                                              hex_val='112233445566778899AA')),
    (SCHEMA_constrained_octetstring(lb=10, ub=22), 'B0112233445566778899AABBCCDDEEFF001122334455', DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=22),
                                                                                                                    hex_val='112233445566778899AABBCCDDEEFF001122334455')),
])
def test_constrained_different_len_octet_string_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=0), '00', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='')),
    (SCHEMA_constrained_ext_octetstring(lb=1, ub=1), '5000', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=1, ub=1), hex_val='A')),
    (SCHEMA_constrained_ext_octetstring(lb=1, ub=1), '5580', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=1, ub=1), hex_val='AB')),
    (SCHEMA_constrained_ext_octetstring(lb=5, ub=5), '001122334450', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=5, ub=5), hex_val='112233445')),
    (SCHEMA_constrained_ext_octetstring(lb=5, ub=5), '001122334455', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=5, ub=5), hex_val='1122334455')),
    (SCHEMA_constrained_ext_octetstring(lb=15, ub=15), '00112233445566778899001122334455', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=15, ub=15),
                                                                                                            hex_val='112233445566778899001122334455')),
    (SCHEMA_constrained_ext_octetstring(lb=22, ub=22), '0011223344556677889900112233445566778899001122', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=22, ub=22),
                                                                                                                          hex_val='11223344556677889900112233445566778899001122')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=2), '00', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=2), '20A0', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='A')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=2), '20AB', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='AB')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=2), '40ABC0', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='ABC')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=2), '40ABCD', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='ABCD')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=15), '08A0', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15), hex_val='A')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=15), '08AB', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15), hex_val='AB')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=15), '48112233445566778899', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15),
                                                                                               hex_val='112233445566778899')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=15), '78112233445566778899AABBCCDDEEFF', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15),
                                                                                                           hex_val='112233445566778899AABBCCDDEEFF')),
])
def test_constrained_with_extension_when_len_is_within_extension_root_octetstring_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=0), '8001A0', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='A')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=0), '8001AB', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='AB')),
    (SCHEMA_constrained_ext_octetstring(lb=0, ub=0), '80051122334455', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='1122334455')),
    (SCHEMA_constrained_ext_octetstring(lb=2, ub=2), '8000', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='')),
    (SCHEMA_constrained_ext_octetstring(lb=2, ub=2), '8001A0', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='A')),
    (SCHEMA_constrained_ext_octetstring(lb=2, ub=2), '8001AB', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='AB')),
    (SCHEMA_constrained_ext_octetstring(lb=2, ub=2), '8003112233', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='112233')),
    (SCHEMA_constrained_ext_octetstring(lb=2, ub=2), '80051122334455', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='1122334455')),
    (SCHEMA_constrained_ext_octetstring(lb=3, ub=4), '8000', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='')),
    (SCHEMA_constrained_ext_octetstring(lb=3, ub=4), '8001A0', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='A')),
    (SCHEMA_constrained_ext_octetstring(lb=3, ub=4), '8001AB', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='AB')),
    (SCHEMA_constrained_ext_octetstring(lb=3, ub=4), '80021122', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='1122')),
    (SCHEMA_constrained_ext_octetstring(lb=3, ub=4), '80051122334455', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='1122334455')),
    (SCHEMA_constrained_ext_octetstring(lb=3, ub=4), '800711223344556677', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='11223344556677')),
    (SCHEMA_constrained_ext_octetstring(lb=10, ub=22), '8009112233445566778899', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=10, ub=22), hex_val='112233445566778899')),
    (SCHEMA_constrained_ext_octetstring(lb=10, ub=22), '80171122334455667788990011223344556677889900112233', DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=10, ub=22),
                                                                                                                              hex_val='1122334455667788990011223344556677889900112233')),
])
def test_constrained_with_extension_when_len_is_NOT_within_extension_root_octetstring_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value



from ..type import SimpleType
from asn1PERser.classes.templates.creator import TemplateFiller


class UTF8String(SimpleType):
    def __init__(self):
        super(UTF8String, self).__init__()
        self.typereference = "UTF8String"

    def fill_template(self):
        return TemplateFiller.fill(asn_type=self.__class__.__name__,
                                   class_name=self.template_class_name)



import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_mixed_schemas import DATA_first_sequence, MySeq, \
    DATA_seq__of__seq_of__of__seq_in_ext_add, DATA_seq__of__seq_of__of__seq_in_ext_add_group, \
    DATA_seq__of__seq_of__of__seq_in_root, OuterSeq, OuterSeq1, OuterSeq2


def test_first_sequence_can_be_decoded():
    assert per_decoder(per_stream=bytearray.fromhex('8004F401640202DEAD02BEEF'), asn1Spec=MySeq()) == DATA_first_sequence()


def test_seq_of_seq_of_of_seq_in_root_can_be_decoded():
    assert per_decoder(per_stream=bytearray.fromhex('80010A0114'), asn1Spec=OuterSeq()) == DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=True, my_sof2_is=False)
    assert per_decoder(per_stream=bytearray.fromhex('C0010A011400011E0128'), asn1Spec=OuterSeq()) == DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=True, my_sof2_is=True)
    assert per_decoder(per_stream=bytearray.fromhex('40011E0128'), asn1Spec=OuterSeq()) == DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=False, my_sof2_is=True)


def test_seq_of_seq_of_of_seq_in_extension_addition_can_be_decoded():
    assert per_decoder(per_stream=bytearray.fromhex('8001F603000500010A0114'), asn1Spec=OuterSeq1()) == DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=True, my_sof2_is=False)
    assert per_decoder(per_stream=bytearray.fromhex('8001F603800500010A01140500011E0128'), asn1Spec=OuterSeq1()) == DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=True, my_sof2_is=True)
    assert per_decoder(per_stream=bytearray.fromhex('8001F602800500011E0128'), asn1Spec=OuterSeq1()) == DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=False, my_sof2_is=True)
    assert per_decoder(per_stream=bytearray.fromhex('0001F6'), asn1Spec=OuterSeq1()) == DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=False, my_sof2_is=False)


def test_seq_of_seq_of_of_seq_in_extension_addition_group_can_be_encoded():
    assert per_decoder(per_stream=bytearray.fromhex('8001F6010580010A0114'), asn1Spec=OuterSeq2()) == DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=True, my_sof2_is=False)
    assert per_decoder(per_stream=bytearray.fromhex('8001F6010AC0010A011400011E0128'), asn1Spec=OuterSeq2()) == DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=True, my_sof2_is=True)
    assert per_decoder(per_stream=bytearray.fromhex('8001F6010540011E0128'), asn1Spec=OuterSeq2()) == DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=False, my_sof2_is=True)
    assert per_decoder(per_stream=bytearray.fromhex('0001F6'), asn1Spec=OuterSeq2()) == DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=False, my_sof2_is=False)


from .creator import template_filler


class Template(object):
    def fill_template(self):
        return template_filler.fill('header')

    @property
    def template_class_name(self):
        return self.identifier.replace('-','_') if self.identifier else self.typereference.replace('-','_')

    @property
    def template_field_type(self):
        return self.typereference.replace('-','_') if self.typereference else self.identifier.replace('-','_')



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class Meters(IntegerType):
    pass


class Picture(OctetStringType):
    pass


class MoveOrder(ChoiceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('up', Meters()),
        NamedType('down', Meters()),
        NamedType('left', Meters()),
        NamedType('right', Meters()),
    )
    componentType = rootComponent


class Data(SequenceType):
    class pictures(SequenceOfType):
        componentType = Picture()

    rootComponent = AdditiveNamedTypes(
        NamedType('id', IntegerType()),
        NamedType('pictures', pictures()),
    )
    componentType = rootComponent


class Order(ChoiceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('move', MoveOrder()),
        NamedType('sendData', Data()),
    )
    componentType = rootComponent





import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_enumerated import SCHEMA_my_enum, \
    SCHEMA_my_ext_enum, short_enum, enumeration_list, DATA_my_enum


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2]), '00',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2]), 'e0')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2]), '80',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2]), 'e1')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:10]), '90',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:10]), 'e9')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:17]), '48',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:17]), 'e9')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33]), '24',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33]), 'e9')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33]), '80',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33]), 'e32')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:100]), 'C4',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:100]), 'e98')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), '7E',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), 'e126')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), '7F',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), 'e127')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), '80',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), 'e128')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), '0080',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e128')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), '00FE',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e254')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), '00FF',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e255')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), '0100',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e256')),
])
def test_no_extension_marker_enumerated_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2], extensionMarker_value=True), '00',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2], extensionMarker_value=True), 'e0')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2], extensionMarker_value=True), '40',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2], extensionMarker_value=True), 'e1')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:10], extensionMarker_value=True), '48',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:10], extensionMarker_value=True), 'e9')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:17], extensionMarker_value=True), '24',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:17], extensionMarker_value=True), 'e9')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33], extensionMarker_value=True), '12',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33], extensionMarker_value=True), 'e9')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33], extensionMarker_value=True), '40',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33], extensionMarker_value=True), 'e32')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:100], extensionMarker_value=True), '62',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:100], extensionMarker_value=True), 'e98')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), '3F00',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), 'e126')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), '3F80',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), 'e127')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), '4000',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), 'e128')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), '000080',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e128')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), '0000FE',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e254')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), '0000FF',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e255')),
    (SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), '000100',
        DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e256')),
])
def test_extension_marker_is_present_and_extension_addition_is_empty_but_value_is_from_root_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:2], extensionAddition_list=short_enum, extensionMarker_value=True), '00',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:2], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e0')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:2], extensionAddition_list=short_enum, extensionMarker_value=True), '40',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:2], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e1')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:10], extensionAddition_list=short_enum, extensionMarker_value=True), '48',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:10], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e9')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:17], extensionAddition_list=short_enum, extensionMarker_value=True), '24',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:17], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e9')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:33], extensionAddition_list=short_enum, extensionMarker_value=True), '12',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:33], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e9')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:33], extensionAddition_list=short_enum, extensionMarker_value=True), '40',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:33], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e32')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:100], extensionAddition_list=short_enum, extensionMarker_value=True), '62',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:100], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e98')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum, extensionMarker_value=True), '3F00',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e126')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum, extensionMarker_value=True), '3F80',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e127')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum, extensionMarker_value=True), '4000',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e128')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), '000080',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e128')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), '0000FE',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e254')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), '0000FF',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e255')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), '000100',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum,
                                        extensionMarker_value=True), 'e256')),
])
def test_extension_marker_is_present_and_extension_addition_is_not_empty_but_value_is_from_root_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:2], extensionMarker_value=True), '80',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:2],
                                        extensionMarker_value=True), 'e0')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:2], extensionMarker_value=True), '81',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:2],
                                        extensionMarker_value=True), 'e1')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:10], extensionMarker_value=True), '89',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:10],
                                        extensionMarker_value=True), 'e9')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:17], extensionMarker_value=True), '89',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:17],
                                        extensionMarker_value=True), 'e9')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:33], extensionMarker_value=True), '89',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:33],
                                        extensionMarker_value=True), 'e9')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:33], extensionMarker_value=True), 'A0',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:33],
                                        extensionMarker_value=True), 'e32')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:100], extensionMarker_value=True), 'C00162',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:100],
                                        extensionMarker_value=True), 'e98')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130], extensionMarker_value=True), 'C0017E',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130],
                                        extensionMarker_value=True), 'e126')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130], extensionMarker_value=True), 'C0017F',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130],
                                        extensionMarker_value=True), 'e127')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130], extensionMarker_value=True), 'C00180',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130],
                                        extensionMarker_value=True), 'e128')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'C00180',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260],
                                        extensionMarker_value=True), 'e128')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'C001FE',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260],
                                        extensionMarker_value=True), 'e254')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'C001FF',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260],
                                        extensionMarker_value=True), 'e255')),
    (SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'C0020100',
        DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260],
                                        extensionMarker_value=True), 'e256')),
])
def test_extension_marker_is_present_and_value_is_from_extension_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value



from pyasn1.type.namedtype import NamedTypes
from asn1PERser.classes.templates import Template
from asn1PERser.classes.types.constraint import NoConstraint


class Type(Template):
    def __init__(self):
        self._typereference = ''
        self._identifier = ''
        self._valuereference = ''
        self._Value = None
        self._optional = False
        self._default = None
        self._constraint = NoConstraint()

    def fill_template(self):
        raise NotImplemented

    def parse_value(self, value):
        return value

    def is_constrained(self):
        if isinstance(self.constraint, NoConstraint):
            return False
        return True

    @property
    def typereference(self):
        return self._typereference

    @typereference.setter
    def typereference(self, typereference):
        self._typereference = typereference

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, identifier):
        self._identifier = identifier

    @property
    def valuereference(self):
        return self._valuereference

    @valuereference.setter
    def valuereference(self, valuereference):
        self._valuereference = valuereference.replace('-', '_')

    @property
    def Value(self):
        return self._Value

    @Value.setter
    def Value(self, value):
        self._Value = self.parse_value(value)

    @property
    def optional(self):
        return self._optional

    @optional.setter
    def optional(self, optional):
        self._optional = optional

    @property
    def default(self):
        if self._default:
            return self.parse_value(self._default)
        return self._default

    @default.setter
    def default(self, default):
        self._default = default

    @property
    def constraint(self):
        return self._constraint

    @constraint.setter
    def constraint(self, constrain):
        self._constraint = constrain

    def __repr__(self):
        type = 'TYPE: {}'.format(self.__class__.__name__)
        typereference = 'typeref: {}'.format(self.typereference)
        identifier = 'id: {}'.format(self.identifier if self.identifier else None)
        valuereference = 'valueref: {}'.format(self.valuereference if self.valuereference else None)
        value = 'val: {}'.format(str(self.Value))
        constrain = 'con: {}'.format(str(self.constraint))
        optional = 'OPT: {}'.format(str(self.optional))
        default = 'DEF: {}'.format(self.default)
        return '(' + ', '.join([type, typereference, identifier, valuereference,
                                value, constrain, optional, default]) + ')'


class SimpleType(Type):
    pass


class ComponentType(Type):
    pass


class DefType(Type):
    pass


class AdditiveNamedTypes(NamedTypes):
    def __init__(self, *named_types):
        self.additive_named_types = list(named_types)
        super(AdditiveNamedTypes, self).__init__(*self.additive_named_types)

    def __add__(self, other):
        if isinstance(other, list):
            all_named_types = []
            for additive_named_type_element in other:
                all_named_types.extend(additive_named_type_element.additive_named_types)
            return AdditiveNamedTypes(*(self.additive_named_types + all_named_types))
        else:
            return AdditiveNamedTypes(*(self.additive_named_types + other.additive_named_types))


import pytest
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.classes.data.builtin.BitStringType import BitStringType
from asn1PERser.classes.types.constraint import ValueSize


def SCHEMA_no_constrain_bitstring():
    class MyBitString(BitStringType):
        '''
        MyBitString ::= BIT STRING
        '''
        pass
    return MyBitString


def SCHEMA_constrained_bitstring(lb, ub):
    class MyBitString(BitStringType):
        '''
        MyBitString ::= BIT STRING (SIZE(lowerEndpoint..upperEndpoint))
        '''
        subtypeSpec = ValueSize(lb, ub)
    return MyBitString


def SCHEMA_constrained_ext_bitstring(lb, ub):
    class MyBitString(BitStringType):
        '''
        MyBitString ::= BIT STRING (SIZE(lowerEndpoint..upperEndpoint, ...))
        '''
        subtypeSpec = ValueSize(lb, ub, extensionMarker=True)
    return MyBitString


def DATA_bitstring(schema, bin_val=None, hex_val=None):
    if bin_val is not None:
        bitstring = schema(binValue=bin_val)
    elif hex_val is not None:
        bitstring = schema(hexValue=hex_val)
    return bitstring


@pytest.mark.parametrize("bitstring, encoded", [
    # (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val=''), '00'),    # Does not work
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='0'), '0100'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='1'), '0180'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='01'), '0240'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='10'), '0280'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='11'), '02C0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='011'), '0360'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='1010'), '04A0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='1111'), '04F0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='111111111111111'), '0FFFFE'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='111111111111111111111111'), '18FFFFFF'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='100000000000000000000001'), '18800001'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='000000000000000000000001'), '18000001'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='0000111000000000010000001111110000000000000001'),
     '2E0E0040FC0004'),
])
def test_no_constrain_bitstring_as_binary_value_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("bitstring, encoded", [
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='0'), '0400'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='1'), '0410'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='10'), '0810'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='01'), '0801'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='A'), '04A0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='0A'), '080A'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='00000ABCDEF'), '2C00000ABCDEF0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF'), '30ABCDEFABCDEF'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF0'), '34ABCDEFABCDEF00'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF0123456789'), '58ABCDEFABCDEF0123456789'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF0123456789ABCDEFABC'),
     '7CABCDEFABCDEF0123456789ABCDEFABC0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF0123456789ABCDEFABC0'),
     '8080ABCDEFABCDEF0123456789ABCDEFABC0'),
    (DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789'),
     '80B0ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789'),
])
def test_no_constrain_bitstring_as_hex_value_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("bitstring, encoded", [
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=0), bin_val=''), '00'),
    # (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=0), hex_val=''), '00'),    # Does not work
])
def test_constrained_bitstring_of_zero_length_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("bitstring, encoded", [
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=1, ub=1), bin_val='1'), '80'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=1, ub=1), bin_val='0'), '00'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=2, ub=2), bin_val='10'), '80'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=2, ub=2), bin_val='11'), 'C0'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=4, ub=4), bin_val='1010'), 'A0'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=4, ub=4), bin_val='0101'), '50'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=7, ub=7), bin_val='1010101'), 'AA'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=7, ub=7), bin_val='0101010'), '54'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=15, ub=15), bin_val='110011001100110'), 'CCCC'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=15, ub=15), bin_val='001100110011001'), '3332'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=16, ub=16), bin_val='1100110011001100'), 'CCCC'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=16, ub=16), bin_val='0011001100110011'), '3333'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=4, ub=4), hex_val='A'), 'A0'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=8, ub=8), hex_val='BB'), 'BB'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=12, ub=12), hex_val='111'), '1110'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=16, ub=16), hex_val='2222'), '2222'),
])
def test_constrained_bitstring_same_length_gt_zero_and_le_16_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("bitstring, encoded", [
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=17, ub=17), bin_val='00110011001100111'), '333380'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=18, ub=18), bin_val='001100110011001101'), '333340'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=19, ub=19), bin_val='0011001100110011001'), '333320'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=20, ub=20), bin_val='00110011001100110011'), '333330'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=21, ub=21), bin_val='001100110011001100111'), '333338'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=22, ub=22), bin_val='0011001100110011001101'), '333334'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=23, ub=23), bin_val='00110011001100110011001'), '333332'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=24, ub=24), bin_val='001100110011001100110011'), '333333'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=25, ub=25), bin_val='0011001100110011001100111'), '33333380'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=128, ub=128), hex_val='ABCDEFABCDEF0123456789ABCDEFABC0'),
     'ABCDEFABCDEF0123456789ABCDEFABC0'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=176, ub=176), hex_val='ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789'),
     'ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789'),

])
def test_constrained_bitstring_same_length_gt_16_and_lt_64K_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("bitstring, encoded", [
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=1), bin_val='1'), '8080'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=15), bin_val='1'), '1080'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=15), bin_val='111000111000111'), 'F0E38E'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=14, ub=15), bin_val='111000111000111'), '80E38E'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=127), bin_val='111000111000111'), '1EE38E'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=1000), bin_val='111000111000111'), '000FE38E'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=10000), bin_val='111000111000111'), '000FE38E'),
    (DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=65535), bin_val='111000111000111'), '000FE38E'),
])
def test_constrained_bitstring_with_ub_less_then_64K_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("bitstring, encoded", [
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=1, ub=1), bin_val='1'), '40'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=1, ub=1), bin_val='0111111'), '80077E'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=2, ub=2), bin_val='10'), '40'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=2, ub=2), bin_val='101'), '8003A0'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=7, ub=7), bin_val='01010101'), '800855'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=15, ub=15), bin_val='11001100110011011'), '8011CCCD80'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=16, ub=16), bin_val='00110011001100110001'), '8014333310'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=25, ub=25), bin_val='0011001100110011001100111111'), '801C333333F0'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=128, ub=128), hex_val='ABCDEFABCDEF0123456789ABCDEFABC0'),
     '00ABCDEFABCDEF0123456789ABCDEFABC0'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=176, ub=176), hex_val='ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789'),
     '00ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=1), bin_val='1'), '4080'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=15), bin_val='1'), '0880'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=15), bin_val='111100001111000'), '78F0F0'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=15), bin_val='0111100001111000'), '80107878'),
    (DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=128), hex_val='ABCDEFABCDEF0123456789ABCDEFABC0'),
     '4000ABCDEFABCDEF0123456789ABCDEFABC0'),
])
def test_constrained_bitstring_with_extension_marker_can_be_encoded(bitstring, encoded):
    assert per_encoder(bitstring) == bytearray.fromhex(encoded)



from collections import namedtuple
from collections import defaultdict
import codecs


__all__ = ['encode']


bit_field_size = namedtuple('bit_field_size', 'octets bits')
bit_field = namedtuple('bit_field', 'octet_align value')


class PerEncodingException(Exception):
    pass


class MandatoryExtensionFieldNotPresent(PerEncodingException):
    pass


class MandatoryFieldNotPresentInExtensionGroup(PerEncodingException):
    pass


class SizeConstrainViolated(PerEncodingException):
    pass


class InvalidComponentIndexIntoStructuredType(PerEncodingException):
    pass


def encode(asn1Spec):
    field_list = []
    asn1Spec.fill_field_list(field_list)
    complete_encoding = _produce_complete_encoding(field_list)
    return codecs.decode(_bin_to_hex(complete_encoding), 'hex_codec')


def _produce_complete_encoding(field_list):    # 11.1
    complete_encoding = ''
    for field in field_list:
        if field.octet_align:
            if len(complete_encoding) % 8 != 0:
                complete_encoding += ''.zfill(abs(len(complete_encoding) % 8 - 8))
            complete_encoding += field.value
            continue
        complete_encoding += field.value
        continue
    if (len(complete_encoding) % 8 != 0) or (len(complete_encoding) == 0):
        complete_encoding += ''.zfill(abs(len(complete_encoding) % 8 - 8))
    return complete_encoding


def _bin_to_hex(binary_str):
    dec_value = int(binary_str, 2)
    return "{value:0{byte_len}x}".format(value=dec_value, byte_len=int(len(binary_str) / 4)).upper()


def encode_boolean(boolean):    # 12
    if boolean._value == True:
        return [bit_field(octet_align=False, value='1')]
    elif boolean._value == False:
        return [bit_field(octet_align=False, value='0')]


def encode_integer(integer):
    if integer.subtypeSpec.extensionMarker:    # 13.1
        if integer.subtypeSpec.lowerEndpoint <= integer._value <= integer.subtypeSpec.upperEndpoint:
            constrained_whole_number_field_list = create_field_list_for_constrained_whole_number(n=integer._value,
                                                                                                 lowerEndpoint=integer.subtypeSpec.lowerEndpoint,
                                                                                                 upperEndpoint=integer.subtypeSpec.upperEndpoint)
            extension_bit_field = bit_field(octet_align=False, value='0')
            return [extension_bit_field] + constrained_whole_number_field_list
        else:
            extension_bit_field = bit_field(octet_align=False, value='1')
            unconstrained_whole_number_field_list = create_field_list_for_unconstrained_whole_number(integer)
            return [extension_bit_field] + unconstrained_whole_number_field_list
    else:
        if is_constant(integer):    # 13.2.1
            pass
        elif is_constrained_whole_number(integer):    # 13.2.2
            constrained_whole_number_field_list = create_field_list_for_constrained_whole_number(n=integer._value,
                                                                                                 lowerEndpoint=integer.subtypeSpec.lowerEndpoint,
                                                                                                 upperEndpoint=integer.subtypeSpec.upperEndpoint)
            return constrained_whole_number_field_list
        elif is_semi_constrained_whole_number(integer):    # 13.2.3
            semi_constrained_whole_number = encode_semi_constrained_whole_number(n=integer._value,
                                                                                 lower_bound=integer.subtypeSpec.lowerEndpoint)
            n = int(len(semi_constrained_whole_number.value) / 8)
            unconstrained_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                         constrained=False,
                                                                         n=n,
                                                                         lowerBound=1,
                                                                         upperBound=None)
            return unconstrained_length_determinant + [semi_constrained_whole_number]

        else:    # 13.2.4
            unconstrained_whole_number_field_list = create_field_list_for_unconstrained_whole_number(integer)
            return unconstrained_whole_number_field_list


def encode_enumerated(enumerated):
    if not enumerated.subtypeSpec.extensionMarker:    # 14.2
        return create_field_list_for_constrained_whole_number(n=enumerated._value, lowerEndpoint=0,
                                                              upperEndpoint=len(list(enumerated.namedValues)) - 1)
    else:    # 14.3
        if enumerated._value in list(enumerated.enumerationRoot.values()):
            extension_bit_field = bit_field(octet_align=False, value='0')
            constrained_whole_number_field_list = create_field_list_for_constrained_whole_number(n=enumerated._value,
                                                                                                 lowerEndpoint=0,
                                                                                                 upperEndpoint=len(list(enumerated.enumerationRoot)) - 1)
            return [extension_bit_field] + constrained_whole_number_field_list
        else:
            n = list(enumerated.extensionAddition.values()).index(enumerated._value)
            extension_bit_field = bit_field(octet_align=False, value='1')
            normally_small_non_negative = encode_normally_small_non_negative_whole_number(n=n, lower_bound=0)
            return [extension_bit_field] + normally_small_non_negative


def encode_bitstring(bitstring):    # 16
    if bitstring.namedValues:    # 16.2, 16.3
        raise NotImplemented

    bitstring_value = bitstring.asBinary()
    bitstring_len = len(bitstring_value)
    extension_bit_field = None
    if bitstring.subtypeSpec.extensionMarker:    # 16.6
        if bitstring.subtypeSpec.lowerEndpoint <= bitstring_len <= bitstring.subtypeSpec.upperEndpoint:
            extension_bit_field = bit_field(octet_align=False, value='0')
        else:
            bitstring_bit_field = bit_field(octet_align=True, value=bitstring_value)
            extension_bit_field = bit_field(octet_align=False, value='1')
            bitstring_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                     constrained=False,
                                                                     n=bitstring_len,
                                                                     lowerBound=0,
                                                                     upperBound=None)
            return [extension_bit_field] + bitstring_length_determinant + [bitstring_bit_field]
    # else:    # 16.7
    if bitstring.subtypeSpec.upperEndpoint == 0:    # 16.8
        return
    if (bitstring.subtypeSpec.lowerEndpoint is not None) and (bitstring.subtypeSpec.lowerEndpoint == bitstring.subtypeSpec.upperEndpoint):
        if bitstring_len <= 16:    # 16.9
            bitstring_bit_field = bit_field(octet_align=False, value=bitstring_value)
            if extension_bit_field:
                return [extension_bit_field, bitstring_bit_field]
            return [bitstring_bit_field]
        elif 16 < bitstring_len < 65536:    # 16.10
            bitstring_bit_field = bit_field(octet_align=True, value=bitstring_value)
            if extension_bit_field:
                return [extension_bit_field, bitstring_bit_field]
            return [bitstring_bit_field]
        else:
            raise NotImplemented
    else:    # 16.11
        bitstring_bit_field = bit_field(octet_align=True, value=bitstring_value)
        if bitstring.subtypeSpec.upperEndpoint and bitstring.subtypeSpec.upperEndpoint < 65536:
            bitstring_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                     constrained=True,
                                                                     n=bitstring_len,
                                                                     lowerBound=bitstring.subtypeSpec.lowerEndpoint,
                                                                     upperBound=bitstring.subtypeSpec.upperEndpoint)
            if extension_bit_field:
                return [extension_bit_field] + bitstring_length_determinant + [bitstring_bit_field]
            return bitstring_length_determinant + [bitstring_bit_field]
        else:
            bitstring_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                     constrained=False,
                                                                     n=bitstring_len,
                                                                     lowerBound=0,
                                                                     upperBound=None)
            if extension_bit_field:
                return [extension_bit_field] + bitstring_length_determinant + [bitstring_bit_field]
            return bitstring_length_determinant + [bitstring_bit_field]


def encode_octetstring(octetstring):
    octetstring_value = ''
    for number in octetstring.asNumbers():
        octetstring_value += bin(number)[2:].zfill(8)
    octetstring_len = int(len(octetstring_value) / 8)

    extension_bit_field = None
    if octetstring.subtypeSpec.extensionMarker:    # 17.3
        if octetstring.subtypeSpec.lowerEndpoint <= octetstring_len <= octetstring.subtypeSpec.upperEndpoint:
            extension_bit_field = bit_field(octet_align=False, value='0')
        else:
            octetstring_bit_field = bit_field(octet_align=False, value=octetstring_value)
            extension_bit_field = bit_field(octet_align=False, value='1')
            octetstring_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                       constrained=False,
                                                                       n=octetstring_len,
                                                                       lowerBound=octetstring.subtypeSpec.lowerEndpoint,
                                                                       upperBound=None)
            return [extension_bit_field] + octetstring_length_determinant + [octetstring_bit_field]
    if octetstring.subtypeSpec.upperEndpoint == 0:    # 17.5
        return
    if (octetstring.subtypeSpec.lowerEndpoint is not None) and (octetstring.subtypeSpec.lowerEndpoint == octetstring.subtypeSpec.upperEndpoint):
        if octetstring_len <= 2:    # 17.6
            octetstring_bit_field = bit_field(octet_align=False, value=octetstring_value)
            if extension_bit_field:
                return [extension_bit_field, octetstring_bit_field]
            return [octetstring_bit_field]
        elif 2 < octetstring_len < 65536:    # 17.7
            octetstring_bit_field = bit_field(octet_align=True, value=octetstring_value)
            if extension_bit_field:
                return [extension_bit_field, octetstring_bit_field]
            return [octetstring_bit_field]
        else:
            raise NotImplemented
    else:    # 17.8
        octetstring_bit_field = bit_field(octet_align=True, value=octetstring_value)
        if octetstring.subtypeSpec.upperEndpoint:
            octetstring_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                       constrained=True,
                                                                       n=octetstring_len,
                                                                       lowerBound=octetstring.subtypeSpec.lowerEndpoint,
                                                                       upperBound=octetstring.subtypeSpec.upperEndpoint)
            if extension_bit_field:
                return [extension_bit_field] + octetstring_length_determinant + [octetstring_bit_field]
            return octetstring_length_determinant + [octetstring_bit_field]
        else:
            octetstring_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                       constrained=False,
                                                                       n=octetstring_len,
                                                                       lowerBound=0,
                                                                       upperBound=None)
            if extension_bit_field:
                return [extension_bit_field] + octetstring_length_determinant + [octetstring_bit_field]
            return octetstring_length_determinant + [octetstring_bit_field]


def encode_sequence(sequence):    # 19
    sequence_field_list = []
    rootComponentFieldList = []
    optional_default_bit_field_value = ''
    for namedType in sequence.rootComponent.namedTypes:
        componentType = sequence.getComponentByName(namedType.name)
        if namedType.isOptional:    # 19.2
            if componentType.isValue:
                if not hasattr(componentType, '_componentValues'):
                    optional_default_bit_field_value += '1'
                    componentType.fill_field_list(rootComponentFieldList)
                else:
                    if len(componentType._componentValues) > 0:
                        optional_default_bit_field_value += '1'
                        componentType.fill_field_list(rootComponentFieldList)
                    else:
                        optional_default_bit_field_value += '0'
            else:
                optional_default_bit_field_value += '0'
        elif namedType.isDefaulted:    # 19.2
            # if isDefaulted, default value must have been initiated when defining asn1 Schema,
            # if default value is different than value passed when filling schema, assume component
            # is present
            if namedType._NamedType__type._value != componentType._value:    # hacky...:) no other way for now
                optional_default_bit_field_value += '1'
                componentType.fill_field_list(rootComponentFieldList)
            else:
                optional_default_bit_field_value += '0'
        else:
            componentType.fill_field_list(rootComponentFieldList)
    if optional_default_bit_field_value:
        if len(optional_default_bit_field_value) < 65536:    # 19.3
            optional_default_bit_field = bit_field(octet_align=False, value=optional_default_bit_field_value)
            rootComponentFieldList = [optional_default_bit_field] + rootComponentFieldList
        else:
            raise NotImplemented
    if sequence.subtypeSpec.extensionMarker:    # 19.1
        extension_bit = bit_field(octet_align=False, value='0')
        if (not sequence.extensionAddition) and (not sequence.extensionAdditionGroups):
            return [extension_bit] + rootComponentFieldList
        if sequence.extensionAddition or sequence.extensionAdditionGroups:    # 19.7
            extension_addition_bit_field_value = ''
            extension_addition_field_list = []
            encoded_extension_addition_groups = []
            extension_group_components = defaultdict(list)
            if sequence.extensionAddition:
                for index, namedType in enumerate(sequence.extensionAddition.namedTypes, start=1):
                    extensionComponentType = sequence.getComponentByName(namedType.name)
                    if extensionComponentType.isValue:
                        if (index > 1) and (not previous_extensionComponent['present']) and (not previous_extensionComponent['optional']):
                            raise MandatoryExtensionFieldNotPresent(previous_extensionComponent['name'])

                        if not hasattr(extensionComponentType, '_componentValues'):
                            extension_addition_bit_field_value += '1'
                            open_type_field_list = encode_open_type_field(extensionComponentType)
                            extension_addition_field_list.extend(open_type_field_list)
                        else:
                            if len(extensionComponentType._componentValues) > 0:
                                extension_addition_bit_field_value += '1'
                                open_type_field_list = encode_open_type_field(extensionComponentType)
                                extension_addition_field_list.extend(open_type_field_list)
                            else:
                                extension_addition_bit_field_value += '0'
                    else:
                        extension_addition_bit_field_value += '0'
                    previous_extensionComponent = {'name': namedType.name,
                                                   'present': bool(int(extension_addition_bit_field_value[-1])),
                                                   'optional': namedType.isOptional}
            if sequence.extensionAdditionGroups:
                for index_group, extensionAdditionGroup in enumerate(sequence.extensionAdditionGroups, start=1):
                    extension_addition_group_present = False
                    previous_extensionGroupComponent = {}
                    for index, namedType in enumerate(extensionAdditionGroup.namedTypes, start=1):
                        previous_extension_addition_group_component_present = False
                        extensionComponentType = sequence.getComponentByName(namedType.name)
                        if extensionComponentType.isValue:
                            if (index > 1) and (not previous_extensionGroupComponent['present']) and (not previous_extensionGroupComponent['optional']):
                                raise MandatoryFieldNotPresentInExtensionGroup(previous_extensionGroupComponent['name'])
                            if not hasattr(extensionComponentType, '_componentValues'):
                                extension_addition_group_present = True
                                previous_extension_addition_group_component_present = True
                            else:
                                if len(extensionComponentType._componentValues) > 0:
                                    extension_addition_group_present = True
                                    previous_extension_addition_group_component_present = True
                        previous_extensionGroupComponent = {'name': namedType.name,
                                                            'present': previous_extension_addition_group_component_present,
                                                            'optional': namedType.isOptional}
                        extension_group_components[index_group].append(previous_extensionGroupComponent)
                    if not extension_addition_group_present:
                        extension_addition_bit_field_value += '0'
                    else:
                        if (index_group > 1):
                            for previous_extension_group_component in extension_group_components[index_group - 1]:
                                if previous_extension_group_component['present'] or previous_extension_group_component['optional']:
                                    break
                            else:
                                raise MandatoryExtensionFieldNotPresent(extension_group_components[index_group - 1][0]['name'])
                        extension_addition_bit_field_value += '1'
                        group_sequence = create_new_fake_empty_sequence(rootComponent=extensionAdditionGroup)
                        for namedType in extensionAdditionGroup.namedTypes:
                            componentType = sequence.getComponentByName(namedType.name)
                            group_sequence._componentValues.append(componentType)
                        encoded = encode_open_type_field(group_sequence)
                        encoded_extension_addition_groups.extend(encoded)
            extension_addition_bit_field = bit_field(octet_align=False, value=extension_addition_bit_field_value)
            if int(extension_addition_bit_field_value, 2) > 0:    # any extensionAddition is present in encoding
                extension_bit = bit_field(octet_align=False, value='1')
                extension_addition_length_determinant = encode_length_determinant(normally_small_length=True,
                                                                                  constrained=False,
                                                                                  n=len(extension_addition_bit_field_value),
                                                                                  lowerBound=None,
                                                                                  upperBound=None)
                sequence_field_list.extend(
                    [extension_bit] + rootComponentFieldList + extension_addition_length_determinant +
                     [extension_addition_bit_field] + extension_addition_field_list + encoded_extension_addition_groups)
            else:
                extension_bit = bit_field(octet_align=False, value='0')
                sequence_field_list.extend([extension_bit] + rootComponentFieldList)
        return sequence_field_list
    return rootComponentFieldList


def encode_sequence_of(sequence_of):    # 20
    sequence_of_field_list = []
    n = len(sequence_of)
    for component in sequence_of:
        component.fill_field_list(sequence_of_field_list)
    if sequence_of.subtypeSpec.extensionMarker:    #20.4
        if sequence_of.subtypeSpec.lowerEndpoint <= n <= sequence_of.subtypeSpec.upperEndpoint:
            extension_bit_field = bit_field(octet_align=False, value='0')
            length_determinant = encode_length_determinant(normally_small_length=False, constrained=True,
                                                           n=n, lowerBound=sequence_of.subtypeSpec.lowerEndpoint,
                                                           upperBound=sequence_of.subtypeSpec.upperEndpoint)
            return [extension_bit_field] + length_determinant + sequence_of_field_list
        else:
            extension_bit_field = bit_field(octet_align=False, value='1')
            length_determinant = encode_length_determinant(normally_small_length=False, constrained=False,
                                                           n=n, lowerBound=0,    # not certain but it works lb=0
                                                           upperBound=None)
            return [extension_bit_field] + length_determinant + sequence_of_field_list
    elif n < sequence_of.subtypeSpec.lowerEndpoint:
        raise SizeConstrainViolated("Number of SET OF/SEQUENCE OF {} components = {}".format(sequence_of.__class__, n))
    elif sequence_of.subtypeSpec.upperEndpoint and (n > sequence_of.subtypeSpec.upperEndpoint):
        raise InvalidComponentIndexIntoStructuredType('invalid index: {}'.format(sequence_of.subtypeSpec.upperEndpoint + 1))
    elif (sequence_of.subtypeSpec.lowerEndpoint == sequence_of.subtypeSpec.upperEndpoint) and (sequence_of.subtypeSpec.upperEndpoint < 65536):    # 20.5
        return sequence_of_field_list
    else:    # 20.6
        if sequence_of.subtypeSpec.upperEndpoint:
            length_determinant = encode_length_determinant(normally_small_length=False, constrained=True,
                                                           n=n, lowerBound=sequence_of.subtypeSpec.lowerEndpoint,
                                                           upperBound=sequence_of.subtypeSpec.upperEndpoint)
            return [] + length_determinant + sequence_of_field_list
        else:
            length_determinant = encode_length_determinant(normally_small_length=False, constrained=False,
                                                           n=n, lowerBound=sequence_of.subtypeSpec.lowerEndpoint, upperBound=None)
            return [] + length_determinant + sequence_of_field_list


def encode_choice(choice):    # 23
    choice_field_list = []
    n = len(choice.rootComponent.namedTypes) - 1
    selected_component = choice.getComponent()
    selected_name = choice.getName()
    selected_component.fill_field_list(choice_field_list)
    if not choice.subtypeSpec.extensionMarker:    # 23.6
        if len(choice.rootComponent.namedTypes) == 1:    # 23.4
            return choice_field_list
        for index, namedType in enumerate(choice.rootComponent.namedTypes):    # 23.6
            if namedType.name == selected_name:
                encoded_choice_index_bit_field = create_field_list_for_constrained_whole_number(n=index,
                                                                                                lowerEndpoint=0,
                                                                                                upperEndpoint=n)
                return [] + encoded_choice_index_bit_field + choice_field_list
    else:    # 23.5
        for index, namedType in enumerate(choice.rootComponent.namedTypes):
            if namedType.name == selected_name:    # 23.7
                extension_bit_bit_field = bit_field(octet_align=False, value='0')
                if len(choice.rootComponent.namedTypes) == 1:
                    return [extension_bit_bit_field] + choice_field_list
                else:
                    encoded_choice_index_bit_field = create_field_list_for_constrained_whole_number(n=index,
                                                                                                    lowerEndpoint=0,
                                                                                                    upperEndpoint=n)
                    return [extension_bit_bit_field] + encoded_choice_index_bit_field + choice_field_list
        else:
            if choice.extensionAddition and not choice.extensionAdditionGroups:
                choice_extension = choice.extensionAddition
            elif not choice.extensionAddition and choice.extensionAdditionGroups:
                choice_extension = choice.extensionAdditionGroups[0].__class__(*[named_type for additive_named_types in choice.extensionAdditionGroups for
                                                                                 named_type in additive_named_types._NamedTypes__namedTypes])
            elif choice.extensionAddition and choice.extensionAdditionGroups:
                choice_extension = choice.extensionAddition + choice.extensionAdditionGroups
            for index, namedType in enumerate(choice_extension.namedTypes):
                if namedType.name == selected_name:
                    extension_bit_bit_field = bit_field(octet_align=False, value='1')
                    encoded_choice_index_bit_field = encode_normally_small_non_negative_whole_number(n=index,
                                                                                                     lower_bound=0)
                    choice_field_list = encode_open_type_field(type=selected_component)
                    return [extension_bit_bit_field] + encoded_choice_index_bit_field + choice_field_list


def create_field_list_for_constrained_whole_number(n, lowerEndpoint, upperEndpoint):
    constrained_whole_number = encode_constrained_whole_number(n=n,
                                                               lowerBound=lowerEndpoint,
                                                               upperBound=upperEndpoint)
    if integer_range(lowerEndpoint, upperEndpoint) <= 65536:  # 13.2.5
        return [constrained_whole_number]
    else:  # 13.2.6
        n = int(len(constrained_whole_number.value) / 8)
        upperBound = get_range_octet_len(upperEndpoint - lowerEndpoint)
        constrained_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                   constrained=True,
                                                                   n=n,
                                                                   lowerBound=1,
                                                                   upperBound=upperBound)
        return constrained_length_determinant + [constrained_whole_number]


def encode_open_type_field(type):    # 11.2
    type_field_list = []
    type.fill_field_list(type_field_list)
    octet_string = _produce_complete_encoding(type_field_list)
    octet_string_len = int(len(octet_string) / 8)
    unconstrained_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                 constrained=False, n=octet_string_len,
                                                                 lowerBound=None, upperBound=None)
    octet_string_bit_field = bit_field(octet_align=True, value=octet_string)
    return unconstrained_length_determinant + [octet_string_bit_field]


def encode_constrained_whole_number(n, lowerBound, upperBound):    # 11.5
    range = upperBound - lowerBound + 1
    bit_field_size = get_bit_field_size(range)
    value = n - lowerBound
    if range == 1:    # 11.5.4
        return bit_field(octet_align=False, value='')
    elif range <= 255:    # 11.5.7.1
        non_negative_binary_integer = encode_non_negative_binary_integer(value, bit_field_size, octet_align=False)
        return non_negative_binary_integer
    elif range == 256:    # 11.5.7.2
        non_negative_binary_integer = encode_non_negative_binary_integer(value, bit_field_size, octet_align=True)
        return non_negative_binary_integer
    elif 257 <= range <= 65536:    # 11.5.7.3
        non_negative_binary_integer = encode_non_negative_binary_integer(value, bit_field_size, octet_align=True)
        return non_negative_binary_integer
    elif range > 65536:    # 11.5.7.4
        non_negative_binary_integer = encode_non_negative_binary_integer(value, bit_field_size,
                                                                         octet_align=True,
                                                                         minimum_number_of_octets=True)
        return non_negative_binary_integer


def encode_normally_small_non_negative_whole_number(n, lower_bound):    # 11.6
    if n <= 63:    # 11.6.1
        single_bit_field = bit_field(octet_align=False, value='0')
        normally_small_non_negative_whole_number = encode_non_negative_binary_integer(value=n,
                                                                                      size=bit_field_size(octets=0, bits=6),
                                                                                      octet_align=False)
        return [single_bit_field, normally_small_non_negative_whole_number]
    else:    # 11.6.2
        single_bit_field = bit_field(octet_align=False, value='1')
        normally_small_non_negative_whole_number = encode_semi_constrained_whole_number(n=n, lower_bound=lower_bound)
        octet_len = int(len(normally_small_non_negative_whole_number.value) / 8)
        unconstrained_length_determinant = encode_length_determinant(normally_small_length=False,
                                                                     constrained=False,
                                                                     n=octet_len,
                                                                     lowerBound=1,
                                                                     upperBound=None)

        return [single_bit_field] + unconstrained_length_determinant + [normally_small_non_negative_whole_number]


def encode_non_negative_binary_integer(value, size, octet_align, minimum_number_of_octets=False):    # 11.3
    if minimum_number_of_octets:    # 11.3.6
        encoded = bin(value)[2:]
        if len(encoded) % 8 != 0:
            encoded = ''.zfill(abs(len(encoded)%8-8)) + encoded
        return bit_field(octet_align=octet_align, value=encoded)
    else:
        encoded = '{value:0{bit_len}b}'.format(value=value, bit_len=size.octets*8 + size.bits)
        return bit_field(octet_align=octet_align, value=encoded)


def encode_semi_constrained_whole_number(n, lower_bound):    # 11.7
    value = n - lower_bound
    semi_constrained_whole_number = encode_non_negative_binary_integer(value=value, size=None, octet_align=True,
                                                                       minimum_number_of_octets=True)
    return semi_constrained_whole_number


def create_field_list_for_unconstrained_whole_number(integer):
    unconstrained_length_determinant = get_twos_complement_octet_len(integer)
    unconstrained_whole_number = encode_unconstrained_whole_number(integer, unconstrained_length_determinant)
    unconstrained_length_determinant = '{0:08b}'.format(unconstrained_length_determinant)
    return [bit_field(octet_align=True, value=unconstrained_length_determinant + unconstrained_whole_number.value)]


def encode_unconstrained_whole_number(integer, unconstrained_length_determinant):    # 11.8
    unconstrained_whole_number = encode_twos_complement_binary_integer(integer,
                                                                       width=8*unconstrained_length_determinant)
    return bit_field(octet_align=True, value=unconstrained_whole_number)


def encode_twos_complement_binary_integer(integer, width):
    if integer._value == 0:
        return '0' * width
    elif integer._value > 0:
        return bin(integer._value)[2:].zfill(width)
    else:
        xored = (abs(integer._value) - 1) ^ int('1' * width, 2)
        return bin(xored)[2:]


def encode_length_determinant(normally_small_length, constrained, n, lowerBound, upperBound):    # 11.9
    if normally_small_length:    # 11.9.3.4
        if n <= 64:
            single_bit_field = bit_field(octet_align=False, value='0')
            normally_small_length_determinant_size = bit_field_size(octets=0, bits=6)
            normally_small_length_determinant = encode_non_negative_binary_integer(value=n - 1,
                                                                                   size=normally_small_length_determinant_size,
                                                                                   octet_align=False)
            return [single_bit_field, normally_small_length_determinant]
        else:
            raise NotImplemented
    elif constrained:    # 11.9.3.3
        constrained_length_determinant = encode_constrained_whole_number(n=n, lowerBound=lowerBound, upperBound=upperBound)
        return [constrained_length_determinant]
    else:    # 11.9.3.5
        if n <= 127:    # 11.9.3.6
            determinant_size = bit_field_size(octets=0, bits=7)
            encoded_length_determinant = encode_non_negative_binary_integer(value=n, size=determinant_size,
                                                                            octet_align=True)
            encoded_length_determinant_with_8th_bit_zero = bit_field(octet_align=encoded_length_determinant.octet_align,
                                                                     value= '0' + encoded_length_determinant.value)
            return [encoded_length_determinant_with_8th_bit_zero]
        elif 127 < n <= 16384:    # 11.9.3.7
            determinant_size = bit_field_size(octets=1, bits=6)
            encoded_length_determinant = encode_non_negative_binary_integer(value=n, size=determinant_size,
                                                                            octet_align=True)
            encoded_length_determinant_with_8th_bit_one_7th_bit_zero = bit_field(octet_align=encoded_length_determinant.octet_align,
                                                                                 value='10' + encoded_length_determinant.value)
            return [encoded_length_determinant_with_8th_bit_one_7th_bit_zero]
        else:
            raise NotImplemented


def get_bit_field_size(range):
    if range == 1:
        return bit_field_size(octets=0, bits=0)
    elif range == 2:
        return bit_field_size(octets=0, bits=1)
    elif 3 <= range <= 4:
        return bit_field_size(octets=0, bits=2)
    elif 5 <= range <= 8:
        return bit_field_size(octets=0, bits=3)
    elif 9 <= range <= 16:
        return bit_field_size(octets=0, bits=4)
    elif 17 <= range <= 32:
        return bit_field_size(octets=0, bits=5)
    elif 33 <= range <= 64:
        return bit_field_size(octets=0, bits=6)
    elif 65 <= range <= 128:
        return bit_field_size(octets=0, bits=7)
    elif 129 <= range <= 255:
        return bit_field_size(octets=0, bits=8)
    elif range == 256:
        return bit_field_size(octets=1, bits=0)
    elif 256 < range <= 65536:
        return bit_field_size(octets=2, bits=0)
    elif range > 65536:
        range_octet_len = get_range_octet_len(range)
        return bit_field_size(octets=range_octet_len, bits=0)


def integer_range(lb, ub):
    return ub - lb + 1


def get_range_octet_len(range):
    base_power = 8
    power_multiplier = 3
    while True:
        if range < 2 ** ((power_multiplier * base_power)):
            return power_multiplier
        else:
            power_multiplier += 1


def get_twos_complement_octet_len(integer):
    base_power = 8
    power_multiplier = 1
    while True:
        if integer._value >= 0:
            if integer._value < 2 ** ((power_multiplier * base_power) - 1):
                return power_multiplier
            else:
                power_multiplier += 1
        else:
            if integer._value >= -(2 ** ((power_multiplier * base_power - 1))):
                return power_multiplier
            else:
                power_multiplier += 1


def is_constrained_whole_number(integer):
    if (integer.subtypeSpec.lowerEndpoint is not None) and (integer.subtypeSpec.upperEndpoint is not None):
        return True
    return False


def is_semi_constrained_whole_number(integer):
    if (integer.subtypeSpec.lowerEndpoint is not None) and (integer.subtypeSpec.upperEndpoint is None):
        return True
    return False


def is_constant(integer):
    return False


def create_new_fake_empty_sequence(rootComponent):
    class FooConstraint(object):
        pass

    class FakeSequence(object):
        subtypeSpec = FooConstraint()
        def __init__(self):
            self.named_types = {}

        def fill_field_list(self, field_list):
            sequence_field_list = encode_sequence(self)
            if sequence_field_list:
                field_list.extend(sequence_field_list)

        def getComponentByName(self, name):
            idx = self.rootComponent.getPositionByName(name)
            return self._componentValues[idx]

        def __setitem__(self, key, value):
            self.named_types[key] = value

    FakeSequence._componentValues = []
    FakeSequence.subtypeSpec.extensionMarker = False
    FakeSequence.subtypeSpec.lowerEndpoint= None
    FakeSequence.subtypeSpec.upperEndpoint= None
    FakeSequence.rootComponent = rootComponent
    FakeSequence.componentType = rootComponent
    FakeSequence.extensionAddition = None
    FakeSequence.extensionAdditionGroups = []
    return FakeSequence()



from pyparsing import Or, Literal, nestedExpr, Optional, OneOrMore, Forward
from .value_def import Value
# from .type_def import Type
from .common_def import SignedNumber, NamedNumber, NamedNumberList, DefinedValue, NULL
from .lexical_items import identifier, LEFT_CURLY_BRACKET, RIGHT_CURLY_BRACKET, \
    COMMA, leftversionbrackets, rightversionbrackets, empty, \
    LEFT_PARENTHESIS, RIGHT_PARENTHESIS, EXCLAMATION_MARK, COLON, \
    CIRCUMFLEX_ACCENT, VERTICAL_LINE, LESS_THAN_SIGN, rangeseparator, \
    ellipsis, number, LEFT_SQUARE_BRACKET, RIGHT_SQUARE_BRACKET, \
    encodingreference, modulereference, typereference, FULL_STOP


INTERSECTION = Literal("INTERSECTION")
UNION = Literal("UNION")
EXCEPT = Literal("EXCEPT")
ALL = Literal("ALL")
MAX = Literal("MAX")
MIN = Literal("MIN")
SIZE = Literal("SIZE")
FROM = Literal("FROM")


ElementSetSpec = Forward()
ElementSetSpecs = Forward()
Constraint = Forward()


# -------------------------------------------
# 51 Subtype elements
# 51.4.4
UpperEndValue = Or([
    Value,
    MAX
])


# 51.4.4
LowerEndValue = Or([
    Value,
    MIN
])

# 51.4.3
UpperEndpoint = Or([
    UpperEndValue,
    LESS_THAN_SIGN + UpperEndValue
])


# 51.4.3
LowerEndpoint = Or([
    LowerEndValue,
    LowerEndValue + LESS_THAN_SIGN
])


# 51.2.1
SingleValue = Value


# 51.4.1
ValueRange = LowerEndpoint + rangeseparator + UpperEndpoint


# 51.5.1
SizeConstraint = SIZE + Constraint


# 51.6.1
# TypeConstraint = Type


# 51.7.1
PermittedAlphabet = FROM + Constraint


# 51.1
SubtypeElements = Or([
    SingleValue,
    # ContainedSubtype,
    ValueRange,
    # PermittedAlphabet,
    SizeConstraint,
    # TypeConstraint,
    # InnerTypeConstraints,
    # PatternConstraint,
    # PropertySettings,
    # DurationRange,
    # TimePointRange,
    # RecurrenceRange
])

# -------------------------------------------

# 50.5
Elements = Or([
    SubtypeElements,
    # ObjectSetElements,
    LEFT_PARENTHESIS + ElementSetSpec + RIGHT_PARENTHESIS
])


# 50.1
IntersectionMark = Or([
    CIRCUMFLEX_ACCENT,
    INTERSECTION
])


# 50.1
UnionMark = Or([
    VERTICAL_LINE,
    UNION
])


# 50.1
Exclusions = EXCEPT + Elements


# 50.1
IntersectionElements = Or([
    Elements + Optional(Exclusions),
    # Elems + Exclusions    # ORIGINAL
])


# 50.1
# Elems << Elements


# 50.1
Intersections = Or([
    OneOrMore(IntersectionElements + Optional(IntersectionMark + IntersectionElements)),
    # IntersectionElements,    # ORIGINAL
    # IElems + IntersectionMark + IntersectionElements    # ORIGINAL
])

# 50.1
# IElems << Intersections


# 50.1
Unions = Or([
    OneOrMore(Intersections + Optional(UnionMark)),
    # Intersections,    # ORIGINAL
    # UElems + UnionMark + Intersections    # ORIGINAL
])


# 50.1
# UElems << Unions


# 50.1
ElementSetSpec << Or([
    Unions,
    ALL + Exclusions
])


# 50.1
AdditionalElementSetSpec = ElementSetSpec


# 50.1
RootElementSetSpec = ElementSetSpec


# 50.1
ElementSetSpecs << Or([
    RootElementSetSpec,
    RootElementSetSpec + COMMA + ellipsis,
    RootElementSetSpec + COMMA + ellipsis + COMMA + AdditionalElementSetSpec
])



# 53.4
ExceptionIdentification = Or([
    SignedNumber,
    DefinedValue,
    # Type + COLON + Value
])


# 53.4
ExceptionSpec = Or([
    # EXCLAMATION_MARK + ExceptionIdentification,
    empty
])


# 49.7
SubtypeConstraint = ElementSetSpecs


# 49.6
ConstraintSpec = Or([
    SubtypeConstraint,
    # GeneralConstraint
])


# 49.6
Constraint << LEFT_PARENTHESIS + ConstraintSpec + ExceptionSpec + RIGHT_PARENTHESIS
